/* gc-major.c */

/* le gc Lang & Dupont */
/* specialement optimise pour favoriser le cas stop & copy */

#include "mlvalues.h"
#include "gc.h"
#include "gc_aux.h"
#include "mem_misc.h"
#include "free_list.h"
#include "roots.h"

extern address heap_start, heap_end;

/* variables globales de la free-list */
extern address next_block, free_list_mark;
extern address *last_link;

/* le gc majeur tond l'herbe sous les pieds du gc mineur...
   il faut donc qu'il connaisse quelques variables de celui-ci.
*/
extern address young_end, young_start;
extern value *ref_table [];
extern value **cur_ref, **ref_index;
extern address *scan_young, *copy_young;

/* variables globales du gc */
address copy_ptr, scan_ptr;

address to_start, to_end;
address from_start, from_end;
static address first_after_from_space;

static address *mark_young, *scan_marked_young;

extern unsigned long heap_size, copy_size;

/***********************************************************************/
/*  fonctions d'acces aux structures de donnees                        */
/***********************************************************************/

static address cache_gray [CACHE_GRAY_SIZE];
static address *cache_gray_ptr;
static int cache_gray_overrun;    /* nb d'objets gris hors du cache */

#define GRAYEN(ptr)                                                           \
/* address ptr; */                                                            \
  { Zassert (IS_IN_HEAP (OBJECT (ptr))                                        \
             &&	!IS_INSIDE_FROM_SPACE (OBJECT (ptr),                          \
                                       H_SIZE (ptr) + 1)                      \
	     && !IS_IN_TO_SPACE (OBJECT (ptr)));                              \
    if (IS_WHITE (OBJECT (ptr))){                                             \
      if (H_TAG (ptr) < NO_SCAN_TAG){  /* c'est un noeud */                   \
	/* mettre l'objet dans le cache pour utilisation par find_gray */     \
	if (cache_gray_ptr == &(cache_gray [CACHE_GRAY_SIZE])){               \
	  WHITE_TO_GRAY (OBJECT (ptr));                                       \
	  cache_gray_overrun++;                                               \
	}else{                                                                \
	  /* on le met dans le cache et on le noircit */                      \
	  Zassert (cache_gray_ptr < &(cache_gray [CACHE_GRAY_SIZE]));         \
	  WHITE_TO_BLACK (OBJECT (ptr));                                      \
	  *cache_gray_ptr++ = (ptr);                                          \
	}                                                                     \
      }else{  /* c'est une feuille */                                         \
	WHITE_TO_BLACK (OBJECT (ptr));                                        \
      }                                                                       \
    }                                                                         \
  }

static address *(zone_start [3]), *(zone_end [3]);
static int next_zone;
static address next_start;

address first_in_ms_space ()
{
  address result = heap_start;

  /* Il faut une indirection sur les debuts et fins de zones, car
     first_after_from_space peut changer pendant le balayage du tas.
  */

  if (from_start < to_start){
    zone_start [0] = &from_start;
    zone_end [0] = &first_after_from_space;
    zone_start [1] = &to_start;
    zone_end [1] = &to_end;
  }else{
    zone_start [0] = &to_start;
    zone_end [0] = &to_end;
    zone_start [1] = &from_start;
    zone_end [1] = &first_after_from_space;
  }
  zone_start [2] = &heap_end;
  zone_end [2] = &heap_start;

  next_zone = 0;

  if (result == *(zone_start [0])){
    result = *(zone_end [0]);
    next_zone = 1;
  }
  if (result == *(zone_start [1])){
    result = *(zone_end [1]);
    next_zone = 2;
  }
  if (result == heap_end){
    result = NULL;
  }

  next_start = *(zone_start [next_zone]);

  return result;
}

#define FIND_NEXT_IN_MS_SPACE(ptr, wrap_flag, escape_label)                   \
/* address ptr; int wrap_flag; ?label? escape_label */                        \
  { (ptr) += H_SIZE (ptr) + 1;                                                \
    Zassert ((ptr) <= heap_end);                                              \
    while ((ptr) == next_start){                                              \
      if (wrap_flag ? 1 : (next_zone != 2)){                                  \
	/* sauter a la prochaine composante connexe de ms-space */            \
	(ptr) = *(zone_end [next_zone]);                                      \
	next_zone = (next_zone + 1) % 3;                                      \
	next_start = *(zone_start [next_zone]);                               \
      }else{                                                                  \
	goto escape_label;                                                    \
      }                                                                       \
    }                                                                         \
  }

static address base_gray, scan_gray;

void init_cache_gray ()
{
  cache_gray_ptr = cache_gray;
  cache_gray_overrun = 0;
}

void init_find_all_gray ()
{
  scan_gray = first_in_ms_space ();
}

address find_all_gray () /* trouve un objet non scanne, le "noircit", et le
			renvoie. */
{
  register address temp;

  if (cache_gray_ptr != cache_gray){
    /* objet dans le cache. Il est deja noirci mais pas trace. */
    Zassert (cache_gray_ptr > cache_gray);
    Zassert (IS_BLACK (OBJECT (*(cache_gray_ptr - 1))));
    return *(--cache_gray_ptr);
  }
  while (scan_ptr < copy_ptr){
    /* dans to-space */
    temp = scan_ptr;
    scan_ptr += H_SIZE (scan_ptr) + 1;
    Zassert (scan_ptr <= copy_ptr);
    if (H_TAG (temp) < NO_SCAN_TAG) return temp;
  }
  while (scan_marked_young < mark_young){
    /* chez les jeunes */
    temp = *scan_marked_young++;
    Zassert (scan_marked_young <= mark_young);
    if (H_TAG (temp) < NO_SCAN_TAG)
      return temp;
  }
  if (cache_gray_overrun){        /* s'il y a des objets gris dans ms-space */
    Zassert (scan_gray != NULL);  /* verifie ms-space != vide */
    /* trouver un objet gris dans ms-space */
    while (1){
      if (IS_GRAY (OBJECT (scan_gray))){
	GRAY_TO_BLACK (OBJECT (scan_gray));
	cache_gray_overrun--;
	return scan_gray;
      }
      /* trouver l'objet suivant dans ms-space */
      FIND_NEXT_IN_MS_SPACE (scan_gray, 1, error_wrap);
      Zassert (scan_gray != NULL);
    }
  }
  return NULL;

  error_wrap: fatal_error ("bug dans FIND_NEXT_IN_MS_SPACE");
}

#define FAST_FIND_GRAY(block, size, none_label)                               \
/* address block; blocksize size; ?label? none_label; */                      \
     /* comme find_all_gray, mais ne balaye pas ms-space */                   \
  { if (cache_gray_ptr != cache_gray){                                        \
      /* objet dans le cache. Il est deja noirci mais pas trace. */           \
      Zassert (cache_gray_ptr > cache_gray);                                  \
      Zassert (IS_BLACK (OBJECT (*(cache_gray_ptr - 1))));                    \
      (block) = *(--cache_gray_ptr);                                          \
      (size) = H_SIZE (block) + 1;                                            \
      goto found;                                                             \
    }                                                                         \
    while (scan_ptr < copy_ptr){                                              \
      /* dans to-space */                                                     \
      (block) = scan_ptr;                                                     \
      (size) = H_SIZE (scan_ptr) + 1;                                         \
      scan_ptr += (size);                                                     \
      Zassert (scan_ptr <= copy_ptr);                                         \
      if (H_TAG (block) < NO_SCAN_TAG)                                        \
	goto found;                                                           \
    }                                                                         \
    /* le cache est vide. on cherche dans les jeunes */                       \
    while (scan_marked_young < mark_young){                                   \
      (block) = *scan_marked_young++;                                         \
      (size) = H_SIZE (block) + 1;                                            \
      Zassert (scan_marked_young <= mark_young);                              \
      if (H_TAG (block) < NO_SCAN_TAG)                                        \
        goto found;                                                           \
    }                                                                         \
    goto none_label;                                                          \
  found: ;                                                                    \
  }


/****************************************************************************/
/*  le gc proprement dit                                                    */
/****************************************************************************/

/* marque les blocs de ms-space, copie les blocs de from-space, et met a jour
   les references vers from-space */

#define MARK_OR_COPY(adr, flag_store_refs)                                    \
/* value *adr; int flag_store_refs; */                                        \
  { register address mc_block = ADDRESS (*(adr));                             \
    /* attention a la capture de variable sur mc_block */                     \
    register blocksize i;                                                     \
    /* adr est l'adresse du pointeur a mettre a jour,                         \
       et mc_block est l'adresse de l'objet en question */                    \
                                                                              \
    Zassert (IS_BLOCK (*(adr)));                                              \
    Zassert (!IS_IN_TO_SPACE (OBJECT (mc_block)));                            \
                                                                              \
    if (mc_block >= from_start){                                              \
      i = H_SIZE (mc_block);	/* bug si l'objet n'est pas dans le tas */    \
      if (mc_block + i + 1 > from_end){                                       \
	/* le bloc est dans ms-space apres from_space */                      \
	if (IS_IN_HEAP (OBJECT (mc_block))){                                  \
	  Zassert (i + 1 > 1);                                                \
	  if (mc_block < first_after_from_space){                             \
	    first_after_from_space = mc_block;                                \
	  }                                                                   \
	  GRAYEN (mc_block);	 /* Je sais : c'est pas de l'anglais. */      \
	}                                                                     \
      }else{ /* bloc dans from-space */                                       \
	if (IS_FORWARDED (OBJECT (mc_block))){                                \
	  /* bloc deja copie */                                               \
	  *(adr) = FORWARD_POINTER (OBJECT (mc_block));                       \
	}else{                                                                \
	  /* bloc a copier */                                                 \
	  register value temp = OBJECT (copy_ptr);                            \
                                                                              \
	  Zassert (i + 1 > 1);                                                \
	  *(header *) copy_ptr = *(header *) mc_block; /* copier le header */ \
          copy_ptr += SIZE_HEADER; mc_block += SIZE_HEADER;                   \
	  *copy_ptr++ = *mc_block++;  /* copier le premier champ */           \
	  if (i -= 1){   /* au moins 2 champs ? */                            \
	    do{                                                               \
	      *copy_ptr++ = *mc_block++;                                      \
	    }while (--i);                                                     \
	  }                                                                   \
	  Zassert (ADDRESS (temp) + SIZE (temp) + 1 == copy_ptr);             \
	  /* mettre en place le pointeur d'indirection et la reference */     \
	  PUT_FORWARD_FLAG (*(adr));                                          \
	  FORWARD_POINTER (*(adr)) = temp;                                    \
	  *(adr) = temp;                                                      \
	}                                                                     \
      }                                                                       \
    }else if (mc_block >= heap_start){                                        \
      Zassert (IS_IN_HEAP (OBJECT (mc_block)));                               \
      i = H_SIZE (mc_block);                                                  \
      Zassert (i + 1 > 1);                                                    \
      /* le bloc est dans ms-space avant from_space */                        \
      GRAYEN (mc_block);	       /* Je sais : c'est pas de l'anglais. */\
    }else if (IS_YOUNG (OBJECT (mc_block))){                                  \
      if (IS_FORWARDED (OBJECT (mc_block))){                                  \
	*(adr) = FORWARD_POINTER (OBJECT (mc_block));                         \
        mark_or_copy (adr, flag_store_refs); /* on recommence... */           \
      }else{                                                                  \
	if (flag_store_refs){                                                 \
	  if (IS_IN_HEAP (OBJECT (adr))){                                     \
	    /* reference inverse. */                                          \
	    *ref_index++ = (adr);                                             \
	    Zassert (ref_index < &ref_table [REF_TABLE_SIZE]);                \
	  }                                                                   \
	}                                                                     \
	if (IS_WHITE (OBJECT (mc_block))){                                    \
	  WHITE_TO_BLACK (OBJECT (mc_block));                                 \
	  *mark_young++ = mc_block;                                           \
	}                                                                     \
      }                                                                       \
    }                                                                         \
  }

void mark_or_copy (adr, flag_store_refs)
     value *adr;
     int flag_store_refs;
{
  MARK_OR_COPY (adr, flag_store_refs);
}

/* parcours du graphe a partir d'une racine donnee */
/* ou bien finir les objets gris si la racine est NULL */
void trace_from (root)
     value *root;
{
  register address p;
  register blocksize i;

  if (root != NULL){
    Zassert (IS_BLOCK (*root));
    Zassert (!IS_IN_HEAP (OBJECT (root)));
  
   mark_or_copy (root, 0);	/* une racine ne peut pas etre dans le heap. */
  }
  while (1){
    /* trouver un bloc gris et sa taille */
    FAST_FIND_GRAY (p, i, no_more);

    Zassert (IS_IN_HEAP (OBJECT (p)) || IS_YOUNG (OBJECT (p)));
    Zassert (H_TAG (p) < NO_SCAN_TAG);
    Zassert (i > 1);

    p += SIZE_HEADER;			/* header */

    if (IS_BLOCK (*p)){			/* premier champ */
      MARK_OR_COPY (p, 1);
    }
    p++;

    if (i -= 2){
      do{
	if (IS_BLOCK (*p)){
	  MARK_OR_COPY (p, 1);
	}
	p++;
      }while (--i);
    }
  }
 no_more: ;
}

/* mark_or_copy_to_black et trace_from_newold servent pour les objets que
   le gc mineur etait en train de copier.
*/

/************ incomprehensible ****************/

void mark_or_copy_to_black (adr)
     value *adr;
{
  address p;
  blocksize i;

  Zassert (IS_WHITE (*adr));
  mark_or_copy (adr, 0);
  FAST_FIND_GRAY (p, i, none);	/* le cache des objets gris doit avoir une
				   taille > 0 */
  Zassert (OBJECT (p) == *adr);
  return;

 none: fatal_error ("bug de mark_or_copy_to_black");
}

void trace_from_newold (p)
     address p;
{
  blocksize i = H_SIZE (p);

  p += SIZE_HEADER;			/* header */
  do{
    if (IS_BLOCK (*p)){
      MARK_OR_COPY (p, 0);
    }
    p++;
  }while (--i);
  trace_from (NULL);		/* suit les objets gris normaux */
}

/**********************************************/

void major_gc ()
{
  /* pour trouver la vraie fin de from-space */
  first_after_from_space = heap_end;
  if (to_start > from_start){
    first_after_from_space = to_start;
  }

  /* initialiser les variables de copie */
  scan_ptr = copy_ptr = to_start;

  /* initialiser le cache pour les objets gris */
  init_cache_gray ();


  /* pour faire la mise a jour de la table des references inverses
     sous les pieds du gc mineur. */
  ref_index = ref_table;
  cur_ref = ref_table - 1;
 
  /* pour stocker les objets gris des jeunes. */
  scan_marked_young = mark_young = copy_young;

  /* les objets que le gc mineur etait en train de copier doivent etre
     traites comme des racines speciales :
     on les trace sans stocker dans ref_table les references directes
     aux objets jeunes.
  */
  { address *p;
    for (p = scan_young; p < copy_young; p++){
      *(value *) p = OBJECT (*p);
      mark_or_copy_to_black ((value *) p);
      *p = ADDRESS (*(value *) p);
    }
    for (p = scan_young; p < copy_young; p++){
      trace_from_newold (*p);
    }
  }

  /* parcours des racines. */
  ALL_ROOTS (trace_from);

  { /* finir les objets gris qui ne sont pas passes dans le cache. */
    address p;
    blocksize i;

    init_find_all_gray ();
    while ((p = find_all_gray ()) != NULL){

      Zassert (IS_IN_HEAP (OBJECT (p)) || IS_YOUNG (OBJECT (p)));
      Zassert (H_TAG (p) < NO_SCAN_TAG);

      i = H_SIZE (p) + 1;
      Zassert (i > 1);
      p += SIZE_HEADER;			/* header */

      if (IS_BLOCK (*p)){		/* premier champ */
	mark_or_copy (p, 1);
      }
      p++;

      if (i -= 2){
	do{
	  if (IS_BLOCK (*p)){
	    mark_or_copy (p, 1);
	  }
	  p++;
	}while (--i);
      }	  
    }
  }

  /* maintenant, il faut blanchir les jeunes qui sont vivants. */
  { address *p;

    Zassert (scan_marked_young == mark_young);
    for (p = copy_young; p < mark_young; p++){
      Zassert (IS_BLACK (OBJECT (*p)));
      BLACK_TO_WHITE (OBJECT (*p));
    }
  }

  Zassert (copy_ptr == scan_ptr);

  /* la vraie fin du from_space : le premier bloc de ms-space qui est utilise */
  from_end = first_after_from_space;

  new_free_list ();

  /* coaguler le reste de to-space avec from-space si possible */
  if (to_end == from_start){
    from_start = copy_ptr;
    to_end = copy_ptr;
  }else{
    ADD_TO_FREE_LIST (copy_ptr, to_end - copy_ptr);
  }

/**** phase de balayage ****/
  {
    register address current_block = first_in_ms_space ();

    if (current_block != NULL){
      while (true){
	if (IS_WHITE (OBJECT (current_block))){
	  ADD_TO_FREE_LIST (current_block,
			    H_SIZE (current_block) + 1);
	}else{
	  Zassert (IS_BLACK (OBJECT (current_block)));
	  BLACK_TO_WHITE (OBJECT (current_block));
	}
	FIND_NEXT_IN_MS_SPACE (current_block, 0, end_of_heap);
	Zassert (IS_IN_HEAP (OBJECT (current_block)));
      }
      end_of_heap: ;
    }
  }
}
