/* roots.h */

/* squelette de definition des racines du runtime */

#ifndef _roots_
#define _roots_

#include "stacks.h"

extern value extra_root;

extern value pending_signal;

#define ALL_ROOTS(trace_fn)                                                   \
/* void (*trace_fn) (value *root); */                                         \
  {                                                                           \
    register value *sp;                                                       \
    register value v;                                                         \
    register int i;                                                           \
                                                                              \
    /* la pile d'arguments */                                                 \
    for (sp = extern_asp; sp < arg_stack_high; sp++) {                        \
      v = *sp;                                                                \
      if ((v != MARK) && (IS_IN_HEAP (v) || IS_YOUNG (v)) && IS_BLOCK (v)) {  \
	trace_fn (sp);                                                        \
      }                                                                       \
    }                                                                         \
                                                                              \
    /* la pile de retour */                                                   \
    for (sp = extern_rsp; sp < ret_stack_high; ) {                            \
      if (IS_IN_HEAP(((struct return_frame *) sp)->env) ||                    \
          IS_YOUNG(((struct return_frame *) sp)->env)) {                      \
        trace_fn(&(((struct return_frame *) sp)->env));                       \
      }                                                                       \
      i = ((struct return_frame *) sp)->cache_size;                           \
      sp = sp + sizeof(struct return_frame) / sizeof(value);                  \
      while (i > 0) {                                                         \
        v = *sp;                                                              \
        if ((IS_IN_HEAP (v) || IS_YOUNG (v)) && IS_BLOCK (v)) {               \
	    trace_fn (sp);                                                    \
	  }                                                                   \
        sp++;                                                                 \
        i--;                                                                  \
      }                                                                       \
    }                                                                         \
									      \
    /* les racines C */							      \
    {									      \
      value *block;							      \
      for (block = C_roots_head; block != NULL; block = (value *) block [1]){ \
	for (sp = block - (long) block [0]; sp < block; sp++){		      \
	  if ((IS_IN_HEAP (*sp) || IS_YOUNG (*sp)) && IS_BLOCK (*sp)){	      \
	    trace_fn (sp);						      \
	  }								      \
	}								      \
      }									      \
    }									      \
									      \
                                                                              \
    /* les globaux */                                                         \
    if (IS_IN_HEAP(global_data) || IS_YOUNG(global_data))                     \
      trace_fn (&global_data);                                                \
    if ((IS_IN_HEAP(pending_signal) || IS_YOUNG(pending_signal))              \
        && IS_BLOCK (pending_signal))                                         \
      trace_fn (&pending_signal);                                             \
                                                                              \
    if ((IS_IN_HEAP (extra_root) || IS_YOUNG (extra_root))                    \
        && IS_BLOCK (extra_root)){                                            \
      trace_fn (&extra_root);                                                 \
    }                                                                         \
                                                                              \
    /* s'il y a d'autres racines, appeler trace_fn pour chacune. */           \
  }

#define ROOTS_FOR_MINOR(trace_fn)                                             \
/* void (*trace_fn) (value *root); */                                         \
  {                                                                           \
    register value *sp;                                                       \
    register value v;                                                         \
    register int i;                                                           \
                                                                              \
    /* la pile d'arguments */                                                 \
    for (sp = extern_asp; sp < arg_stack_high; sp++) {                        \
      v = *sp;                                                                \
      if ((v != MARK) && IS_YOUNG (v) && IS_BLOCK (v)) {                      \
	trace_fn (sp);                                                        \
      }                                                                       \
    }                                                                         \
                                                                              \
    /* la pile de retour */                                                   \
    for (sp = extern_rsp; sp < ret_stack_high; ) {                            \
      if (IS_YOUNG(((struct return_frame *) sp)->env))                        \
        trace_fn(&(((struct return_frame *) sp)->env));                       \
      i = ((struct return_frame *) sp)->cache_size;                           \
      sp = sp + sizeof(struct return_frame) / sizeof(value);                  \
      while (i > 0) {                                                         \
        v = *sp;                                                              \
        if (IS_YOUNG (v) && IS_BLOCK (v)) {                                   \
	    trace_fn (sp);                                                    \
	  }                                                                   \
        sp++;                                                                 \
        i--;                                                                  \
      }                                                                       \
    }                                                                         \
                                                                              \
    /* les racines C */							      \
    {									      \
      value *block;							      \
      for (block = C_roots_head; block != NULL; block = (value *) block [1]){ \
	for (sp = block - (long) block [0]; sp < block; sp++){		      \
	  if ((IS_IN_HEAP (*sp) || IS_YOUNG (*sp)) && IS_BLOCK (*sp)){	      \
	    trace_fn (sp);						      \
	  }								      \
	}								      \
      }									      \
    }									      \
                                                                              \
    /* Les globaux */                                                         \
    if (IS_YOUNG(global_data))                                                \
      trace_fn(&global_data);                                                 \
    if (IS_YOUNG(pending_signal) && IS_BLOCK (pending_signal))                \
      trace_fn (&pending_signal);                                             \
                                                                              \
    /* s'il y a d'autres racines, appeler trace_fn pour chacune. */           \
  }


#endif /* _roots_ */
