/* Initialisation des piles et de la table des globaux */

#include "mlvalues.h"
#include "stacks.h"
#include "mem_misc.h"

value * arg_stack_low;
value * arg_stack_high;
value * arg_stack_threshold;
value * ret_stack_low;
value * ret_stack_high;
value * ret_stack_threshold;
value * extern_asp;
value * extern_rsp;
struct trap_frame * tp;
value global_data;
value *C_roots_head;

/* Tout preparer pour un GC pendant l'allocation statique */

void reset_roots()
{
  arg_stack_low = 0;
  arg_stack_high = 0;
  extern_asp = 0;
  ret_stack_low = 0;
  ret_stack_high = 0;
  extern_rsp = 0;
  C_roots_head = NULL;
  global_data = (value) 0;    /* Pas MLINT(0), confer ALL_ROOTS... */
  tp = (struct trap_frame *) 0;
}

void init_stacks(arg_stack_size, ret_stack_size)
     blocksize arg_stack_size, ret_stack_size;
{
  arg_stack_low = (value *) stat_alloc(arg_stack_size * sizeof(value));
  arg_stack_high = arg_stack_low + arg_stack_size;
  arg_stack_threshold = arg_stack_low + ARG_STACK_THRESHOLD;
  extern_asp = arg_stack_high;
  ret_stack_low = (value *) stat_alloc(ret_stack_size * sizeof(value));
  ret_stack_high = ret_stack_low + ret_stack_size;
  ret_stack_threshold = ret_stack_low + RET_STACK_THRESHOLD;
  extern_rsp = ret_stack_high;
  tp = (struct trap_frame *) ret_stack_high;
}

static void realloc_arg_stack()
{        
  blocksize new_size;
  value * new_low, * new_high, * new_asp;
  struct trap_frame * p;

  Zassert(extern_asp >= arg_stack_low);
  new_size = arg_stack_high - arg_stack_low + ARG_STACK_GROW;
  gc_message("Growing argument stack to ",
             new_size * sizeof(value) / 1024,
             "k.\n");
  new_low = (value *) stat_alloc(new_size * sizeof(value));
  new_high = new_low + new_size;
  new_asp = new_high - (arg_stack_high - extern_asp);

  bcopy((char *) extern_asp,
        (char *) new_asp,
        (arg_stack_high - extern_asp) * sizeof(value));
  stat_free(arg_stack_low);
  for (p = tp; p < (struct trap_frame *) ret_stack_high; p = p->tp)
    p->asp = new_high - (arg_stack_high - p->asp);
  arg_stack_low = new_low;
  arg_stack_high = new_high;
  arg_stack_threshold = arg_stack_low + ARG_STACK_THRESHOLD;
  extern_asp = new_asp;
}

static void realloc_ret_stack()
{        
  blocksize new_size;
  value * new_low, * new_high, * new_rsp;
  struct trap_frame * p;

  Zassert(extern_rsp >= ret_stack_low);
  new_size = ret_stack_high - ret_stack_low + RET_STACK_GROW;
  gc_message("Growing return stack to ",
             new_size * sizeof(value) / 1024,
             "k.\n");
  new_low = (value *) stat_alloc(new_size * sizeof(value));
  new_high = new_low + new_size;
  new_rsp = new_high - (ret_stack_high - extern_rsp);
  bcopy((char *) extern_rsp,
        (char *) new_rsp,
        (ret_stack_high - extern_rsp) * sizeof(value));
  stat_free(ret_stack_low);
  tp = (struct trap_frame *) (new_high - (ret_stack_high - (value *) tp));
  for (p = tp; p < (struct trap_frame *) new_high; p = p->tp) {
    p->tp = (struct trap_frame *)
      (new_high - (ret_stack_high - (value *) p->tp));
  }
  ret_stack_low = new_low;
  ret_stack_high = new_high;
  ret_stack_threshold = ret_stack_low + RET_STACK_THRESHOLD;
  extern_rsp = new_rsp;
}

void realloc_stacks()
{
  if (extern_rsp < ret_stack_threshold)
    realloc_ret_stack();
  if (extern_asp < arg_stack_threshold)
    realloc_arg_stack();
}

